unit jwWord97 ;

// Original code by Joe Nodeland jnodeland@dccnet.com on 18 December 1999
// Amended by Dr Allan Harkness 24 December 1999

// Joe provided me with a descendent of my classes with some useful table
// methods. I have converted these to basic procedures.
// I chose not to incorporate them into the classes because I feel that
// they fill a specific requirement for some people (database -> table)
// and are unlikely to be useful to everyone (eg I don't use tables and I don't
// use the BDE TQuery).
// Also, from the variety of uses I have seen for my units, everyone seems
// to do things differently. Therefor these procedures are more likely to
// be used as a starting point for you to work from. Even if you don't plan
// to use tables, I would recommend that you read the code as there are quite
// a few good ideas for communicating to/from Word.

// ProcessBookmarks iterates through all the document's bookmarks,
// retrieves each field name and replaces the text with the field value

// ProcessTables:
// Joe's method takes an already existing formatted table and adds each field's
// value one by one, row by row. This has the advantage of having a preformatted
// table (eg widths, fonts, etc).
// This is the method most commonly described in the newsgroups

// ProcessTables2:
// This is my version. It writes out the table as delimited text then
// converts the text to a formatted table. It then does some formatting
// of the table justification (see the macro code).

// As you can see, my method is much faster. Note that the routine will insert
// data in chunks if the table is huge - this avoids concatenating enormous strings
// I have chosen every 20 rows - in this demo it will never require to use more
// than one insertion at the end. Experiment with this number to see what is the
// best for performance for you - depends on the size of the fields.
// Instead of inserting after every 20 rows, you could insert when sTable gets
// to a certain size

interface

uses
  {$IFDEF VER130} {$IFDEF USE_WORD2K} Word2000, {$ELSE} Word97, {$ENDIF} {$ELSE} Word_TLB, {$ENDIF}
  dbTables, db, sysUtils, AHWord97, classes;

procedure ProcessBookmarks (wrdDoc : TWordDoc; qryData :TQuery);
procedure ProcessTables (wrdDoc : TWordDoc; qryData : TQuery);
procedure ProcessTables2 (wrdDoc : TWordDoc; qryData : TQuery);

implementation

Const FIELD              = 'Field ';
      NOT_FOUND          = ' not found ';
      SEP1               = '#';
      SEP2               = '@';
      TABLE              = 'TABLE';

function Token (var sSource : String; Seperator : String) : String;
var
  x : Integer;
begin
  x := Pos (Seperator, sSource);
  if x > 0 then
  begin
    Result := Copy (sSource, 1, x - 1);
    sSource := Copy (sSource, x + 1, MaxInt);
  end
  else
  begin
    Result := sSource;
    sSource := '';
  end;
end;

function FormatField (aField : TField):String;
begin
  case aField.Datatype of
    ftCurrency : Result := Format ('%m', [aField.asCurrency]);
    else         Result := aField.asString;
  end;
end;

procedure ProcessBookmarks (wrdDoc : TWordDoc; qryData : TQuery);
// iterates through the bookmarks in the active document and picks out
// any that are prefixed 'fld' - it replaces each of these with the
// appropriate field value. The field name is the text of the bookmark
// before it is replaced (this allows field names to have spaces)
var
  wrdRange : TWordRange;
  sFieldName, sBookmarkName :String;
  iBookmark : Integer;
begin
  qryData.Open;
  wrdRange := TWordRange.CreateFromSelection (wrdDoc);
  try
    iBookmark := 1;
    while (iBookmark <= wrdDoc.NoOfBookMarks) do
    begin
      sBookmarkName := wrdDoc.BookmarkByIndex [1].Name;
      if Copy (sBookmarkName, 1, 3) = 'fld' then
      begin
        wrdRange.GotoBookmark (sBookmarkName);
        sFieldName := Trim (wrdRange.Text);
        if qryData.FindField (sFieldName) <> nil then
          wrdRange.Text := qryData.FieldByName (sFieldName).asString
        else
          wrdRange.Text := FIELD + sFieldName + NOT_FOUND;
      end
      else // not a field bookmark
        inc (iBookmark)
    end;
  finally
    wrdRange.free;
  end;
end;

procedure ParseTableInfo (sText : String; var iTableId : Integer; slFieldList : TStringList);
// parses info text to get table number, field names and associated column numbers
// eg from 'TABLE#1#Field1@1#Field2@2#Field3@3'
var
  sFieldData : String;
  sFieldName :String;
  iFieldPos : Integer;
begin
  slFieldList.Clear;
  if Token (sText, SEP1) = TABLE then
  begin
    iTableId := StrToInt (Token (sText, SEP1));
    //extract field names and positions from comment and add to list
    While sText <> '' do begin
      sFieldData := Token (sText, SEP1);
      sFieldName := Token (sFieldData, SEP2);
      iFieldPos  := StrToInt (sFieldData);
      slFieldList.AddObject (sFieldName, Pointer (iFieldPos));
    end;
  end
  else
    iTableId := -1; //not valid table info
end;

procedure FillTableWithData (wrdDoc : TWordDoc; var iTableId : Integer;
                             qryData : TQuery; slFieldList : TStringList);
var
  numRows, numCols, mergesplit : OleVariant;
  i, iCol, iColCount, iRowCount :Integer;
  sFieldName, sFieldData :String;
begin
  If not qryData.Active then qryData.Open else qryData.First;
  iColCount := wrdDoc.Document.Tables.Item (iTableId).Columns.count;
  iRowCount := wrdDoc.Document.Tables.Item (iTableId).Rows.count;
  numrows := 2;
  numcols := 1;
  mergeSplit := False;

  with qryData, wrdDoc.Document.Tables.Item (iTableId) do
  while not Eof do
  begin
    Rows.Item (iRowCount).Cells.Split (numRows, numCols, mergesplit);
    inc (iRowCount);
    for i := 0 to slFieldList.Count -1 do begin
      sFieldName := slFieldList [i];
      If FindField (sFieldName) <> nil then
        sFieldData := FormatField (FieldbyName (sFieldName))
      else sFieldData := FIELD + sFieldName + NOT_FOUND;
      if sFieldData = '' then sFieldData := '~';
      iCol := Integer (slFieldList.Objects [i]);
      if iCol <= iColCount then
        Cell (iRowCount, iCol).Range.Text := sFieldData;
    end;
    Next;
  end;
end;

procedure ProcessTables (wrdDoc : TWordDoc; qryData : TQuery);
var
  DocComments : Comments;
  slFieldList : TStringList;
  i : Integer;
  TableId : Integer;
begin
  slFieldList := TStringList.Create;
  try
    DocComments := wrdDoc.Document.Comments;
    for i := 1 to DocComments.Count do
    begin
      ParseTableInfo (DocComments.Item (i).Range.Text, TableId, slFieldList);
      if TableId <> -1 then FillTableWithData (wrdDoc, TableId, qryData, slFieldList);
    end;
  finally
    slFieldList.Free;
  end;
end;

procedure FillTableWithData2 (wrdDoc : TWordDoc;
                              sBookMarkName, sMacroName : String;
                              qryData : TQuery; slFieldList : TStringList);
// NB expects a field for every column
var
  iCol, iRow, iFieldNo : Integer;
  wrdRange : TWordRange;
  sFieldName, sFieldData, sTable : String;
begin
  If not qryData.Active then qryData.Open else qryData.First;
  sTable := '';
  iRow := 0;
  WrdRange := TWordRange.CreateFromBookMark (wrdDoc, sBookMarkName);
  try
    with qryData, wrdDoc do
    begin
      while not Eof do
      begin
        inc (iRow);
        sTable := sTable + #13;
        for iCol := 1 to slFieldList.Count do
        begin
          iFieldNo := slFieldList.IndexOfObject (TObject (iCol));
          if iFieldNo <> -1 then
          begin
            sFieldName := slFieldList [iFieldNo];
            If FindField (sFieldName) <> nil then
              sFieldData := FormatField (FieldbyName (sFieldName))
            else
              sFieldData := FIELD + sFieldName + NOT_FOUND;
            if sFieldData = '' then sFieldData := '~';
          end
          else
            sFieldData := '~';
          sTable := sTable + sFieldData + '|';
        end;
        System.Delete (sTable, Length (sTable), 1); //trim trailing '|'
        if iRow mod 3 = 0 then // insert chunk of table so that sTable never
        begin                  // gets to be huge
          WrdRange.InsertAfter (sTable);
          sTable := ''
        end;
        Next;
      end;
      if sTable <> '' then WrdRange.InsertAfter (sTable);
      WrdRange.Select;
      WordApp.RunMacro (sMacroName);
    end;
  finally
    WrdRange.Free;
  end;
end;

procedure ProcessTables2 (wrdDoc : TWordDoc; qryData : TQuery);
var
  DocComments : Comments;
  slFieldList : TStringList;
  i : Integer;
  TableId : Integer;
begin
  slFieldList := TStringList.Create;
  try
    DocComments := wrdDoc.Document.Comments;
    for i := 1 to DocComments.Count do
    begin
      ParseTableInfo (DocComments.Item (i).Range.Text, TableId, slFieldList);
      if TableId <> -1 then FillTableWithData2 (wrdDoc,
                                                'tbl' + IntToStr (TableId),
                                                'DataToTable' + IntToStr (TableId),
                                                qryData, slFieldList);
    end;
  finally
    slFieldList.Free;
  end;
end;

end.
